%% Loading and pre-processing the data
clear
load('mnist.mat')

x_train = x_train ./ 255;
x_valid = x_valid ./ 255;
x_test = x_test ./ 255;

y_train = categorical(y_train); length(y_train)
y_valid = categorical(y_valid); length(y_valid)
y_test = categorical(y_test); length(y_test)

histogram(y_train)
imshow(x_train(:, :, :, 1))


%% Defining and training the model

rng(3)
layers = [
    imageInputLayer([28 28 1])
	
    convolution2dLayer(3,16,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,16,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,64,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
    	
    fullyConnectedLayer(5)
    softmaxLayer
    classificationLayer];

miniBatchSize = 32;
options = trainingOptions( 'adam', 'MiniBatchSize', miniBatchSize, ...
    'MaxEpochs', 50, 'Shuffle', 'every-epoch', ...
    'ValidationData', {x_valid, y_valid}, 'ValidationFrequency', 10, ...
    'Plots', 'training-progress');

net = trainNetwork(x_train, y_train, layers, options);

%% Testing the model

y_pred = classify(net, x_test);
test_accuracy = sum(y_pred == y_test) / numel(y_test);
test_accuracy
confusionchart(y_test, y_pred)

%% Data augmentation

rng(3)
[n_per_class, edges] = histcounts(y_train);
classes = [0 1 2 3 4];
max_class = max(n_per_class);
add_per_class = max_class - n_per_class;

idx_added = [];
for c = 1:5
    added = 0;
    while added < add_per_class(c)
        idx_add = randi(size(y_train, 1));
        if y_train(idx_add) == categorical(classes(c))
            idx_added(end + 1) = idx_add;
            added = added + 1;
        end
    end
end

x_train_added = add_noise(x_train(:, :, :, idx_added));
y_train_added = y_train(idx_added);
x_train_aug = cat(4, x_train, x_train_added);
y_train_aug = cat(1, y_train, y_train_added);
histogram(y_train_aug)

%% Training the model with the augmentated dataset

rng(3)
net = trainNetwork(x_train_aug, y_train_aug, layers, options);

y_pred = classify(net, x_test);
test_accuracy = sum(y_pred == y_test) / numel(y_test);

confusionchart(y_test, y_pred)

%% Adding shifting to Augmentation

rng(3)
x_train_added = add_noise(shift(x_train(:, :, :, idx_added)));
x_train_aug = cat(4, x_train, x_train_added);

net = trainNetwork(x_train_aug, y_train_aug, layers, options);

y_pred = classify(net, x_test);
test_accuracy = sum(y_pred == y_test) / numel(y_test); test_accuracy

confusionchart(y_test, y_pred)

%% Reduced number of filters in Conv layers

rng(3)
layers = [
    imageInputLayer([28 28 1])
	
    convolution2dLayer(3,8,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,8,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
	
    convolution2dLayer(3,32,'Padding',1)
    reluLayer
    maxPooling2dLayer(2,'Stride',2)
    	
    fullyConnectedLayer(5)
    softmaxLayer
    classificationLayer];

miniBatchSize = 32;
options = trainingOptions( 'adam', 'MiniBatchSize', miniBatchSize, ...
    'MaxEpochs', 15, 'Shuffle', 'every-epoch', ...
    'ValidationData', {x_valid, y_valid}, 'ValidationFrequency', 10, ...
    'Plots', 'training-progress');

net = trainNetwork(x_train_aug, y_train_aug, layers, options);

y_pred = classify(net, x_test);
test_accuracy = sum(y_pred == y_test) / numel(y_test); test_accuracy

confusionchart(y_test, y_pred)